from cirrocumulus.simple_data import SimpleData
from natsort.natsort import natsorted


class FeatureAggregator:

    def __init__(self, obs_measures=[], var_measures=[], dimensions=[]):
        self.obs_measures = obs_measures
        self.var_measures = var_measures
        self.dimensions = dimensions

    @staticmethod
    def add_to_result(df, result):
        if df is not None:
            measure_df = df.T
            for feature in measure_df:
                values = measure_df[feature]
                feature_result = {'min': float(values.loc['min']),
                                  'max': float(values.loc['max']),
                                  'sum': float(values.loc['sum']),
                                  'mean': float(values.loc['mean'])}
                if 'numExpressed' in values.index:
                    feature_result['numExpressed'] = int(values.loc['numExpressed'])
                result[feature] = feature_result

    def execute(self, adata):
        result = {}
        if adata.shape[0] > 0:
            for column in self.dimensions:
                df = adata.obs.agg({column: lambda x: x.value_counts(sort=False)})
                sorted_categories = natsorted(df.index)
                df = df.iloc[df.index.get_indexer_for(sorted_categories)]
                dimension_summary = {'categories': df.index.to_list(),
                                     'counts': df[column].to_list()}
                result[column] = dimension_summary

            if len(self.var_measures) > 0:
                FeatureAggregator.add_to_result(SimpleData.X_stats(adata, self.var_measures), result)
            if len(self.obs_measures) > 0:
                FeatureAggregator.add_to_result(SimpleData.obs_stats(adata, self.obs_measures), result)
        return result



import anndata
import numpy as np
import pandas as pd
import scipy.sparse

from cirrocumulus.abstract_dataset import AbstractDataset
from cirrocumulus.simple_data import SimpleData

# only works with local files
from cirrocumulus.star_fusion_utils import read_star_fusion_file


class AnndataDataset(AbstractDataset):

    def __init__(self, backed=None, force_sparse=True, extensions=['h5ad', 'loom', 'zarr']):
        super().__init__()
        self.path_to_data = {}
        self.backed = backed
        self.force_sparse = force_sparse
        self.extensions = extensions

    def get_suffixes(self):
        return self.extensions

    def read_adata(self, path):
        path_lc = path.lower()
        if path_lc.endswith('.loom'):
            return anndata.read_loom(path)
        elif path_lc.endswith('.zarr'):
            return anndata.read_zarr(path)
        elif path_lc.endswith('.tsv'):
            return read_star_fusion_file(path)
        return anndata.read(path, backed=self.backed)
        # elif path.endswith('.mtx'):
        #
        #     return anndata.read_mtx(path, backed=self.backed)
        # elif path.endswith('.txt'):
        #
        #     return anndata.read_text(path, backed=self.backed)
        # elif path.endswith('.csv'):
        #
        #     return anndata.read_csv(path, backed=self.backed)

    def add_data(self, path, data):
        self.path_to_data[path] = data

    def get_data(self, path):
        adata = self.path_to_data.get(path)
        if adata is None:
            adata = self.read_adata(path)
            if scipy.sparse.isspmatrix_csr(adata.X):
                adata.X = adata.X.tocsc()
            self.add_data(path, adata)
        return adata

    def schema(self, file_system, path):
        return SimpleData.schema(self.get_data(path))

    def read(self, file_system, path, obs_keys=[], var_keys=[], basis=None, dataset=None, schema=None):
        adata = self.get_data(path)
        obs = None
        X = None

        if len(var_keys) > 0:
            X = adata[:, var_keys].X
            if len(X.shape) == 1:
                X = np.array([X]).T
            if self.force_sparse and not scipy.sparse.issparse(X):
                X = scipy.sparse.csr_matrix(X)

        for key in obs_keys:
            if key == 'index':
                values = adata.obs.index.values
            else:
                values = adata.obs[key].values
            if obs is None:
                obs = pd.DataFrame()
            obs[key] = values

        if basis is not None and len(basis) > 0:
            if obs is None:
                obs = pd.DataFrame()
            for b in basis:
                embedding_name = b['name']
                embedding_data = adata.obsm[embedding_name]
                dimensions = b['dimensions']
                for i in range(dimensions):
                    obs[b['coordinate_columns'][i]] = embedding_data[:, i]
        return SimpleData(X, obs, pd.DataFrame(index=pd.Index(var_keys)))



class DatabaseAPI:

    def __init__(self):
        self.provider = None

    def server(self):
        return self.provider.server()

    def user(self, email):
        return self.provider.user(email)

    def datasets(self, email):
        return self.provider.datasets(email)

    def delete_dataset(self, email, dataset_id):
        return self.provider.delete_dataset(email, dataset_id)

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        return self.provider.get_dataset(email, dataset_id, ensure_owner)

    def upsert_dataset(self, email, dataset_id, dataset_name, url, readers):
        return self.provider.upsert_dataset(email, dataset_id, dataset_name, url, readers)

    def dataset_filters(self, email, dataset_id):
        return self.provider.dataset_filters(email, dataset_id)

    def get_dataset_filter(self, email, filter_id):
        return self.provider.get_dataset_filter(email, filter_id)

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        return self.provider.upsert_dataset_filter(email, dataset_id, filter_id, filter_name, filter_notes,
            dataset_filter)

    def delete_dataset_filter(self, email, filter_id):
        return self.provider.delete_dataset_filter(email, filter_id)



import numpy as np
import scipy.sparse

from cirrocumulus.simple_data import SimpleData


def get_basis(basis, nbins=None, agg=None, dimensions=2, precomputed=False):
    if isinstance(dimensions, str):
        dimensions = int(dimensions)
    coordinate_columns = []
    for i in range(dimensions):
        coordinate_columns.append(basis + '_' + str(i + 1))
    full_name = basis + '_' + str(dimensions)
    if nbins is not None:
        full_name = full_name + '_' + str(nbins) + '_' + str(agg)
    return {'name': basis, 'dimensions': dimensions, 'coordinate_columns': coordinate_columns, 'nbins': nbins,
            'agg': agg, 'full_name': full_name, 'precomputed': precomputed}


class EmbeddingAggregator:

    def __init__(self, obs_measures, var_measures, dimensions, count, nbins, basis, agg_function, quick=False):
        self.nbins = nbins
        self.agg_function = agg_function
        self.obs_measures = obs_measures
        self.var_measures = var_measures
        self.dimensions = dimensions
        self.add_count = count
        self.basis = basis
        self.quick = quick  # do not compute purity. convert X to dense

    @staticmethod
    def convert_coords_to_bin(df, nbins, coordinate_columns, bin_name, coordinate_column_to_range=None):
        # replace coordinates with bin, set df[name] to bin
        for name in coordinate_columns:
            values = df[name].values
            if coordinate_column_to_range is not None:
                view_column_range = coordinate_column_to_range[name]
                column_min = view_column_range[0]
                column_max = view_column_range[1]
            else:
                column_min = values.min()
                column_max = values.max()
            df[name] = np.floor(np.interp(values, [column_min, column_max], [0, nbins - 1])).astype(int)
        if len(coordinate_columns) == 2:
            df[bin_name] = df[coordinate_columns[0]] * nbins + df[coordinate_columns[1]]
        else:
            df[bin_name] = df[coordinate_columns[2]] + nbins * (
                    df[coordinate_columns[1]] + nbins * df[coordinate_columns[0]])

    def execute(self, adata):
        result = {'coordinates': {}, 'values': {}}
        var_measures = self.var_measures
        obs_measures = self.obs_measures
        dimensions = self.dimensions
        add_count = self.add_count
        basis = self.basis
        nbins = self.nbins
        agg_function = self.agg_function
        quick = self.quick
        if nbins is not None:
            df = adata.obs
            if add_count:
                df['__count'] = 1.0

            # bin level summary, coordinates have already been converted
            agg_dict = {}
            full_basis_name = basis['full_name']
            for column in basis['coordinate_columns']:
                agg_dict[column] = 'min'
            for column in obs_measures:
                agg_dict[column] = agg_function
            if add_count:
                agg_dict['__count'] = 'sum'
            grouped = df.groupby(full_basis_name)

            is_X_sparse = False
            has_var_measures = len(var_measures) > 0
            has_dimensions = len(dimensions) > 0
            if has_var_measures:
                X = adata.X[:, SimpleData.get_var_indices(adata, var_measures)]
                is_X_sparse = scipy.sparse.issparse(X)
            if quick or not is_X_sparse:
                for i in range(len(var_measures)):
                    df[var_measures[i]] = X[:, i] if not is_X_sparse else X[:, i].toarray().flatten()
                    agg_dict[var_measures[i]] = agg_function
            if quick:
                def mode(x):
                    return x.mode()[0]

                for dimension in dimensions:
                    # skip purity
                    agg_dict[dimension] = mode

            df_summary = grouped.agg(agg_dict)
            X_output = None
            if not quick:

                dimension_purity_output = {}
                dimension_mode_output = {}
                for column in dimensions:
                    dimension_purity_output[column] = []
                    dimension_mode_output[column] = []
                if (has_var_measures and is_X_sparse) or len(dimensions) > 0:
                    for key, g in grouped:
                        indices = grouped.indices[key]
                        if has_dimensions:
                            group_df = df.iloc[indices]
                            for dimension in dimensions:
                                value_counts = group_df[dimension].value_counts(sort=False)
                                largest = value_counts.nlargest(1)
                                purity = largest[0] / value_counts.sum()
                                dimension_mode_output[dimension].append(largest.index[0])
                                dimension_purity_output[dimension].append(purity)
                        if has_var_measures and is_X_sparse:
                            X_group = X[indices]
                            if agg_function == 'max':
                                X_summary = X_group.max(axis=0)
                                X_summary = X_summary.toarray().flatten()
                            elif agg_function == 'min':
                                X_summary = X_group.min(axis=0)
                                X_summary = X_summary.toarray().flatten()
                            elif agg_function == 'mean':
                                X_summary = X_group.mean(axis=0)
                                X_summary = X_summary.A1
                            elif agg_function == 'sum':
                                X_summary = X_group.sum(axis=0)
                                X_summary = X_summary.A1
                            X_output = np.vstack((X_output, X_summary)) if X_output is not None else X_summary

            for i in range(len(var_measures)):
                if X_output is not None:
                    result['values'][var_measures[i]] = X_output[:, i]
                else:
                    result['values'][var_measures[i]] = df_summary[var_measures[i]]
            for i in range(len(obs_measures)):
                result['values'][obs_measures[i]] = df_summary[obs_measures[i]]
            if add_count:
                result['values']['__count'] = df_summary['__count']

            for column in dimensions:
                if not quick:
                    result['values'][column] = dict(value=dimension_mode_output[column],
                        purity=dimension_purity_output[column])
                else:
                    result['values'][column] = dict(value=df_summary[column])
            result['bins'] = df_summary.index
            for column in basis['coordinate_columns']:
                result['coordinates'][column] = df_summary[column]
        else:  # no binning
            if add_count:
                result['values']['__count'] = np.ones(adata.shape[0])
            if len(var_measures) > 0:
                X = adata.X[:, SimpleData.get_var_indices(adata, var_measures)]
                is_X_sparse = scipy.sparse.issparse(X)
                for i in range(len(var_measures)):
                    result['values'][var_measures[i]] = X[:, i] if not is_X_sparse else X[:,
                                                                                        i].toarray().flatten()
            for column in obs_measures + dimensions:
                result['values'][column] = adata.obs[column]
            for column in basis['coordinate_columns']:
                result['coordinates'][column] = adata.obs[column]
        return result



import numpy as np
import pandas as pd
import scipy.sparse


class SimpleData:

    def __init__(self, X, obs, var):

        self.obs = obs
        self.var = var
        if X is not None:
            if len(X.shape) == 1:
                X = np.array([X]).T
            n_var = X.shape[1]
            n_obs = X.shape[0]
        else:
            n_var = len(var) if var is not None else 0
            n_obs = len(obs) if obs is not None else 0
        self.X = X
        self.shape = (n_obs, n_var)

    @staticmethod
    def view(adata, row_slice):
        X = adata.X[row_slice] if adata.X is not None else None
        obs = adata.obs[row_slice] if adata.obs is not None else None
        return SimpleData(X, obs, adata.var)


    @staticmethod
    def obs_stats(adata, columns):
        df = adata.obs[columns]
        # variables on columns, stats on rows, transpose so that stats are on columns
        return df.agg(['min', 'max', 'sum', 'mean']).T

    @staticmethod
    def X_stats(adata, var_ids):
        indices = SimpleData.get_var_indices(adata, var_ids)

        X = adata.X[:, indices]
        min_values = X.min(axis=0)
        mean_values = X.mean(axis=0)
        max_values = X.max(axis=0)
        sums = X.sum(axis=0)

        if scipy.sparse.issparse(X):
            min_values = min_values.toarray().flatten()
            max_values = max_values.toarray().flatten()
            mean_values = mean_values.A1
            sums = sums.A1
            num_expressed = X.getnnz(axis=0)
        else:
            num_expressed = (X != 0).sum(axis=0)

        return pd.DataFrame(data={'min': min_values, 'max': max_values, 'sum': sums, 'numExpressed': num_expressed,
                                  'mean': mean_values}, index=var_ids)

    @staticmethod
    def get_var_indices(adata, names):
        return adata.var.index.get_indexer_for(names)

    @staticmethod
    def schema(adata):
        obs_cat = []
        obs = []
        result = {'version': '1.0.0'}
        for key in adata.obs_keys():
            if pd.api.types.is_categorical_dtype(adata.obs[key]) or pd.api.types.is_bool_dtype(adata.obs[key]):
                obs_cat.append(key)
            else:
                obs.append(key)
        spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None

        if spatial_node is not None:
            spatial_node_keys = list(spatial_node.keys())
            if len(spatial_node_keys) == 1:
                spatial_node = spatial_node[spatial_node_keys[0]].keys()  # images', 'metadata', 'scalefactors']

        result['var'] = list(
            sorted(adata.var_names.values, key=lambda x: ('zzzzz' + x.lower()) if x[0].isdigit() else x.lower()))
        result['obs'] = obs
        result['obsCat'] = obs_cat
        result['shape'] = adata.shape
        embeddings = []
        for key in adata.obsm_keys():

            dim = min(3, adata.obsm[key].shape[1])
            embedding = dict(name=key, dimensions=dim)

            if key == 'spatial':
                if spatial_node is not None and 'scalefactors' in spatial_node and 'images' in spatial_node:
                    embedding['spatial'] = dict(scalefactors=spatial_node['scalefactors'],
                        images=list(spatial_node['images'].keys()))
            if dim == 3:
                embeddings.append(embedding)
                embedding = embedding.copy()
                embedding['dimensions'] = 2
                embeddings.append(embedding)
            else:
                embeddings.append(embedding)
        result['embeddings'] = embeddings
        return result

    @staticmethod
    def to_df(adata, obs_measures, var_measures, dimensions, basis=None):
        df = pd.DataFrame()
        obs_keys = obs_measures + dimensions
        if basis is not None:
            obs_keys += basis['coordinate_columns']

        for key in obs_keys:
            df[key] = adata.obs[key]
        indices = SimpleData.get_var_indices(adata, var_measures)
        for i in range(len(var_measures)):
            X = adata.X[:, indices[i]]
            if scipy.sparse.issparse(X):
                X = X.toarray()
            df[var_measures[i]] = X
        return df



import numpy as np
import pandas as pd
import pytest

from cirrocumulus.data_processing import handle_embedding
from cirrocumulus.embedding_aggregator import EmbeddingAggregator, get_basis


def create_df(test_data, measures, dimensions, basis):
    df = pd.DataFrame(test_data[:, measures].X.toarray(), columns=measures)
    df = df.join(test_data.obs[dimensions].reset_index())
    return df.join(pd.DataFrame(test_data.obsm['X_umap'][:, 0:2], columns=basis['coordinate_columns']))


def test_no_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis):
    basis = get_basis(basis)
    results = handle_embedding(dataset_api=dataset_api, dataset=input_dataset, basis=basis,
        measures=measures + list(map(lambda x: 'obs/' + x, continuous_obs)), dimensions=dimensions)

    df = create_df(test_data, measures, dimensions + continuous_obs, basis)
    for key in basis['coordinate_columns']:
        np.testing.assert_array_equal(results['coordinates'][key], df[key].values, err_msg=key)
    for key in measures:
        np.testing.assert_array_equal(results['values'][key], df[key].values, err_msg=key)
    for key in dimensions:
        np.testing.assert_array_equal(results['values'][key], test_data.obs[key].values, err_msg=key)


def group_df(test_data, measures, dimensions, continuous_obs, basis):
    agg_dict = {}
    df = create_df(test_data, measures, dimensions + continuous_obs, basis)
    for key in basis['coordinate_columns']:
        agg_dict[key] = 'max'
    for key in measures + continuous_obs:
        agg_dict[key] = 'max'
    for key in dimensions:
        agg_dict[key] = lambda x: x.mode()[0]
    EmbeddingAggregator.convert_coords_to_bin(df, nbins=100,
        coordinate_columns=basis['coordinate_columns'],
        bin_name=basis['full_name'])
    # group by bin then agg

    return df.groupby(basis['full_name']).agg(agg_dict)


def diff_binning(grouped_df, measures, dimensions, continuous_obs, basis, results):
    for key in basis['coordinate_columns']:
        np.testing.assert_array_equal(results['coordinates'][key], grouped_df[key].values, err_msg=key)
    for key in measures:
        np.testing.assert_array_equal(results['values'][key], grouped_df[key].values, err_msg=key)
    for key in continuous_obs:
        np.testing.assert_array_equal(results['values'][key], grouped_df[key].values, err_msg=key)
    for key in dimensions:
        np.testing.assert_array_equal(results['values'][key]['value'], grouped_df[key].values, err_msg=key)


@pytest.fixture(autouse=True)
def quick():
    return [True, False]


def test_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis, quick):
    nbins = 100
    agg_function = 'max'
    basis = get_basis(basis, nbins, agg_function)
    grouped_df = group_df(test_data, measures, dimensions, continuous_obs, basis)
    results = handle_embedding(dataset_api=dataset_api, dataset=input_dataset,
        basis=basis, measures=measures + list(map(lambda x: 'obs/' + x, continuous_obs)), dimensions=dimensions,
        quick=quick)
    diff_binning(grouped_df, measures, dimensions, continuous_obs, basis, results)


# def test_saved_embedding(tmp_path, dataset_api, test_data, measures, dimensions, continuous_obs, basis):
#     import zarr
#     url = str(tmp_path) + '.zarr'
#     make_ordered(test_data.obs, None)
#     test_data = test_data[:, measures]
#     test_data.write_zarr(url)
#     store = zarr.open(url)
#     nbins = 100
#     agg_function = 'max'
#     basis_name = basis
#     obsm = test_data.obsm[basis_name]
#     basis = get_basis(basis, nbins, agg_function)
#     for i in range(len(basis['coordinate_columns'])):
#         test_data.obs[basis['coordinate_columns'][i]] = obsm[:, i]
#
#     EmbeddingAggregator.convert_coords_to_bin(df=test_data.obs,
#         nbins=basis['nbins'],
#         coordinate_columns=basis['coordinate_columns'],
#         bin_name=basis['full_name'])
#
#     result = EmbeddingAggregator(obs_measures=continuous_obs,
#         var_measures=[], dimensions=dimensions,
#         count=False,
#         nbins=basis['nbins'], basis=basis, agg_function=basis['agg']).execute(test_data)
#     basis_group = require_binned_basis_group(store, basis)
#     obs_group = basis_group.require_group('obs')
#     coords_group = basis_group.require_group('coords')
#     write_basis_obs(basis, coords_group, obs_group, result)
#     result = EmbeddingAggregator(obs_measures=[],
#         var_measures=measures, dimensions=[],
#         count=False,
#         nbins=basis['nbins'], basis=basis, agg_function=basis['agg']).execute(test_data)
#     write_basis_X(coords_group, basis_group, test_data, result)
#
#     input_dataset = Entity(url, {'name': tmp_path, 'url': url})
#     result = dataset_api.read_precomputed_basis(input_dataset, obs_keys=dimensions + continuous_obs, var_keys=measures,
#         basis=basis)
#     grouped_df = group_df(test_data, measures, dimensions, continuous_obs, basis)
#     diff_binning(grouped_df, measures, dimensions, continuous_obs, basis, result)



import json

import cachecontrol
import google.auth.transport.requests
import requests
from flask import request
from google.oauth2 import id_token


class GoogleAuth:

    def __init__(self):
        with open('cirrocumulus/config.json', 'r') as f:
            self.config = json.load(f)
        session = requests.session()
        cached_session = cachecontrol.CacheControl(session)
        self.cached_request = google.auth.transport.requests.Request(session=cached_session)

    @property
    def client_id(self):
        return self.config['clientId']

    def auth(self):
        token = request.headers.get('Authorization')
        if not token.startswith('Bearer '):
            raise ValueError('Token should start with Bearer ')
        token = token.split('Bearer ')[1]
        idinfo = id_token.verify_oauth2_token(token, self.cached_request, self.config['clientId'])
        if idinfo['aud'] != self.config['clientId']:
            raise ValueError('Wrong aud')
        if idinfo['iss'] not in ['accounts.google.com', 'https://accounts.google.com']:
            raise ValueError('Wrong issuer.')
        return idinfo



import numpy as np
import pandas as pd
import scipy.sparse
from cirrocumulus.simple_data import SimpleData
from natsort import natsorted
from pandas import CategoricalDtype


class DotPlotAggregator:

    def __init__(self, var_measures, dimensions):
        self.var_measures = var_measures
        self.dimensions = dimensions

    def execute(self, adata):
        results = []
        # {categories:[], name:'', values:[{name:'', fractionExpressed:0, mean:0}]}
        var_measures = self.var_measures
        dimensions = self.dimensions
        if len(var_measures) == 0 or len(dimensions) == 0:
            return results
        X = adata.X[:, SimpleData.get_var_indices(adata, var_measures)]
        issparse = scipy.sparse.issparse(X)
        df = adata.obs
        for dimension in dimensions:
            if pd.api.types.is_categorical_dtype(df[dimension]):
                if not df[dimension].dtype.ordered:
                    df[dimension] = df[dimension].astype(
                        CategoricalDtype(natsorted(df[dimension].dtype.categories), ordered=True))
                if len(df[dimension].dtype.categories) <= 1:
                    continue
            grouped = adata.obs.groupby(dimension)
            group_names = []
            mean_output = None
            fraction_expressed_output = None
            for key, g in grouped:
                group_names.append(key)
                indices = grouped.indices[key]
                X_group = X[indices]
                mean_values = X_group.mean(axis=0)

                if issparse:
                    mean_values = mean_values.A1
                    fraction_expressed = X_group.getnnz(axis=0) / X_group.shape[0]
                else:
                    fraction_expressed = (X_group != 0).sum(axis=0) / (X_group.shape[0])

                mean_output = np.vstack((mean_output, mean_values)) if mean_output is not None else mean_values
                fraction_expressed_output = np.vstack((fraction_expressed_output,
                                                       fraction_expressed)) if fraction_expressed_output is not None else fraction_expressed

            values = []
            dotplot_result = {'categories': group_names, 'name': dimension, 'values': values}
            for i in range(mean_output.shape[1]):
                name = var_measures[i]
                values.append({'name': name,
                               'fractionExpressed': fraction_expressed_output[:, i],
                               'mean': mean_output[:, i]})
            results.append(dotplot_result)
        return results



from cirrocumulus.dotplot_aggregator import DotPlotAggregator
from cirrocumulus.embedding_aggregator import EmbeddingAggregator, get_basis
from cirrocumulus.feature_aggregator import FeatureAggregator
from cirrocumulus.ids_aggregator import IdsAggregator
from cirrocumulus.simple_data import SimpleData
from cirrocumulus.unique_aggregator import UniqueAggregator


def filter_adata(adata, data_filter):
    keep_expr = get_adata_filter(adata, data_filter)
    adata = SimpleData.view(adata, keep_expr) if keep_expr is not None else adata
    return adata


def get_adata_filter(adata, data_filter):
    import pandas as pd
    import scipy.sparse
    keep_expr = None
    if data_filter is not None:

        user_filters = data_filter.get('filters', [])
        combine_filters = data_filter.get('combine', 'and')
        for filter_obj in user_filters:
            field = filter_obj[0]
            op = filter_obj[1]
            value = filter_obj[2]
            if isinstance(field, dict):  # selection box

                selected_points_basis = get_basis(field['basis'], field.get('nbins'),
                    field.get('agg'), field.get('ndim', '2'), field.get('precomputed', False))

                if 'points' in value:
                    p = value['points']
                    field = selected_points_basis['full_name'] if selected_points_basis[
                                                                      'nbins'] is not None else 'index'
                    if field == 'index':
                        keep = adata.obs.index.isin(p)
                    else:
                        keep = adata.obs[field].isin(p)
                else:
                    keep = None
                    for p in value['path']:
                        if 'z' in p:  # 3d
                            selection_keep = \
                                (adata.obs[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p[
                                    'height']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][2]] >= p['z']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][2]] <= p['z'] + p['depth'])
                        else:
                            selection_keep = \
                                (adata.obs[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                                (adata.obs[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p['height'])

                    keep = selection_keep | keep if keep is not None else selection_keep

            else:
                if adata.obs is not None and field in adata.obs:
                    X = adata.obs[field]
                else:
                    indices = SimpleData.get_var_indices(adata, [field])
                    X = adata.X[:, indices[0]]
                if op == 'in':
                    keep = (X.isin(value)).values
                elif op == '>':
                    keep = (X > value)
                elif op == '=':
                    keep = (X == value)
                elif op == '<':
                    keep = (X < value)
                elif op == '!=':
                    keep = (X != value)
                elif op == '>=':
                    keep = (X >= value)
                elif op == '<=':
                    keep = (X <= value)
                else:
                    raise ValueError('Unknown filter')

            if scipy.sparse.issparse(keep):
                keep = keep.toarray().flatten()
            if isinstance(keep, pd.Series):
                keep = keep.values
            if keep_expr is not None:
                if combine_filters == 'and':
                    keep_expr = keep_expr & keep
                else:
                    keep_expr = keep_expr | keep
            else:
                keep_expr = keep

    return keep_expr


def precomputed_summary(dataset_api, dataset, obs_measures, var_measures, dimensions):
    if '__count' in var_measures:
        var_measures.remove('__count')
    return dataset_api.read_precomputed_stats(dataset, obs_keys=dimensions + obs_measures, var_keys=var_measures)


def precomputed_grouped_stats(dataset_api, dataset, var_measures, dimensions):
    if (len(var_measures)) > 0 and len(dimensions) > 0:
        return dataset_api.read_precomputed_grouped_stats(dataset,
            var_keys=var_measures, obs_keys=dimensions)
    return []


def precomputed_embedding(dataset_api, dataset, basis, obs_measures, var_measures,
                          dimensions):
    if (len(obs_measures) + len(var_measures) + len(dimensions)) == 0:
        obs_measures = ['__count']
    return dataset_api.read_precomputed_basis(dataset, obs_keys=obs_measures + dimensions, var_keys=var_measures,
        basis=basis)


def get_var_name_type(key):
    index = key.find('/')
    if index == -1:
        return key, 'X'
    else:
        key_type = key[0:index]
        name = key[index + 1:]
        return name, key_type


def split_measures(measures):
    obs_measures = []
    var_measures = []
    for i in range(len(measures)):
        name, key_type = get_var_name_type(measures[i])
        if key_type == 'X':
            var_measures.append(name)
        elif key_type == 'obs':
            obs_measures.append(name)
        else:
            raise ValueError('Unknown key type: ' + key_type)
    return obs_measures, var_measures


def handle_embedding(dataset_api, dataset, basis, measures=[], dimensions=[], quick=True):
    obs_measures, var_measures = split_measures(measures)
    if basis['precomputed']:
        result = precomputed_embedding(dataset_api, dataset, basis, obs_measures, var_measures, dimensions)
    else:
        count = '__count' in var_measures
        if count:
            var_measures.remove('__count')
        adata = dataset_api.read(dataset=dataset, obs_keys=dimensions + obs_measures,
            var_keys=var_measures, basis=[basis])
        if basis['nbins'] is not None:
            EmbeddingAggregator.convert_coords_to_bin(df=adata.obs,
                nbins=basis['nbins'],
                coordinate_columns=basis['coordinate_columns'],
                bin_name=basis['full_name'])
        result = EmbeddingAggregator(obs_measures=obs_measures,
            var_measures=var_measures, dimensions=dimensions,
            count=count,
            nbins=basis['nbins'],
            basis=basis,
            agg_function=basis['agg'],
            quick=quick).execute(adata)
    return result


def handle_grouped_stats(dataset_api, dataset, measures=[], dimensions=[]):
    # all dot plot measures are in X
    result = {}
    if dataset_api.has_precomputed_stats(dataset):
        result['dotplot'] = precomputed_grouped_stats(dataset_api, dataset, measures, dimensions)
    else:
        adata = dataset_api.read(dataset=dataset, obs_keys=dimensions, var_keys=measures)
        result['dotplot'] = DotPlotAggregator(var_measures=measures,
            dimensions=dimensions).execute(adata)
    return result


def handle_stats(dataset_api, dataset, measures=[], dimensions=[]):
    result = {}
    obs_measures, var_measures = split_measures(measures)
    if dataset_api.has_precomputed_stats(dataset):
        result['summary'] = precomputed_summary(dataset_api, dataset, obs_measures, var_measures,
            dimensions)
    else:
        adata = dataset_api.read(dataset=dataset, obs_keys=obs_measures + dimensions, var_keys=var_measures)
        result['summary'] = FeatureAggregator(obs_measures, var_measures, dimensions).execute(adata)
    return result


def handle_export_dataset_filters(dataset_api, dataset, data_filters):
    import json
    import pandas as pd

    reformatted_filters = []
    filter_names = []
    for data_filter_obj in data_filters:
        filter_value = json.loads(data_filter_obj['value'])
        filter_names.append(data_filter_obj['name'])
        reformatted_filters.append(filter_value)

    adata_info = get_data(dataset_api, dataset, measures=['obs/index'], data_filters=reformatted_filters)
    result_df = pd.DataFrame(index=adata_info['adata'].obs['index'])
    for i in range(len(reformatted_filters)):
        data_filter = reformatted_filters[i]
        filter_name = filter_names[i]
        adata_filtered = filter_adata(adata_info['adata'], data_filter)
        df = pd.DataFrame(index=adata_filtered.obs['index'])
        df[filter_name] = True
        result_df = result_df.join(df, rsuffix='r')
    result_df.fillna(False, inplace=True)
    return result_df.to_csv()


def handle_diff_exp(dataset_api, dataset, data_filter, var_range):
    adata_info = get_data(dataset_api, dataset, [], [], [], [data_filter])
    adata = adata_info['adata']
    mask = get_adata_filter(adata, data_filter)
    return dataset_api.diff_exp(dataset, mask, var_range)


def handle_selection(dataset_api, dataset, embeddings=[], measures=[], dimensions=[], data_filter=None, stats=True):
    selection_info = get_selected_data(dataset_api, dataset, embeddings=embeddings, measures=measures,
        dimensions=dimensions, data_filter=data_filter)
    basis_objs = selection_info['basis']
    obs_measures = selection_info['obs_measures']
    var_measures = selection_info['var_measures']
    adata = selection_info['adata']

    result = {}
    if len(basis_objs) > 0:
        result['coordinates'] = {}
        for basis_obj in basis_objs:
            result['coordinates'][basis_obj['full_name']] = UniqueAggregator(
                basis_obj['full_name'] if basis_obj['nbins'] is not None else 'index').execute(adata)
    if stats:
        result['summary'] = FeatureAggregator(obs_measures, var_measures, dimensions).execute(adata)
    result['count'] = adata.shape[0]
    return result


def handle_selection_ids(dataset_api, dataset, data_filter):
    selection_info = get_selected_data(dataset_api, dataset, measures=['obs/index'], data_filter=data_filter)
    return {'ids': IdsAggregator().execute(selection_info['adata'])}


def get_data(dataset_api, dataset, embeddings=[], measures=[], dimensions=[], data_filters=[]):
    obs_measures, var_measures = split_measures(measures)
    var_keys_filter = []
    obs_keys_filter = []
    selected_points_filter_basis_list = []
    for data_filter in data_filters:
        _var_keys_filter, _obs_keys_filter, selected_points_filter_basis_list = data_filter_keys(data_filter)
        selected_points_filter_basis_list += selected_points_filter_basis_list
        var_keys_filter += _var_keys_filter
        obs_keys_filter += _obs_keys_filter
    obs_keys = list(set(obs_measures + obs_keys_filter + dimensions))
    var_keys = list(set(var_measures + var_keys_filter))
    basis_objs = []
    basis_keys = set()
    for embedding in embeddings:
        basis_obj = get_basis(embedding['basis'], embedding.get('nbins'), embedding.get('agg'),
            embedding.get('ndim', '2'), embedding.get('precomputed', False))
        if basis_obj['full_name'] not in basis_keys:
            basis_objs.append(basis_obj)
            basis_keys.add(basis_obj['full_name'])
    for basis_obj in selected_points_filter_basis_list:
        if basis_obj['full_name'] not in basis_keys:
            basis_objs.append(basis_obj)
            basis_keys.add(basis_obj['full_name'])
    adata = dataset_api.read(dataset=dataset, obs_keys=obs_keys, var_keys=var_keys, basis=basis_objs)
    for basis_obj in basis_objs:
        if not basis_obj['precomputed'] and basis_obj['nbins'] is not None:
            EmbeddingAggregator.convert_coords_to_bin(df=adata.obs,
                nbins=basis_obj['nbins'],
                bin_name=basis_obj['full_name'],
                coordinate_columns=basis_obj['coordinate_columns'])
    return {'adata': adata, 'basis': basis_objs, 'obs_measures': obs_measures, 'var_measures': var_measures,
            'dimensions': dimensions}


def get_selected_data(dataset_api, dataset, embeddings=[], measures=[], dimensions=[], data_filter=None):
    adata_info = get_data(dataset_api, dataset, embeddings, measures, dimensions,
        [data_filter] if data_filter is not None else [])
    adata = filter_adata(adata_info['adata'], data_filter)
    adata_info['adata'] = adata
    return adata_info


def data_filter_keys(data_filter):
    var_keys = set()
    obs_keys = set()
    basis_list = []
    if data_filter is not None:
        user_filters = data_filter.get('filters', [])

        for i in range(len(user_filters)):
            user_filter = user_filters[i]
            key = user_filter[0]
            if isinstance(key, dict):
                basis = get_basis(key['basis'], key.get('nbins'), key.get('agg'),
                    key.get('ndim', '2'), key.get('precomputed', False))
                basis_list.append(basis)
            else:
                name, key_type = get_var_name_type(key)
                user_filter[0] = name
                if key_type == 'X':
                    var_keys.add(name)
                else:
                    obs_keys.add(name)
    return list(var_keys), list(obs_keys), basis_list



import json
import os

from cirrocumulus.entity import Entity


def create_dataset_meta(path):
    result = {'id': path, 'url': path, 'name': os.path.splitext(os.path.basename(path))[0]}
    if os.path.basename(path).endswith('.json'):
        with open(path, 'rt') as f:
            result.update(json.load(f))
    return result


class LocalDbAPI:

    def __init__(self, path):
        self.path = path
        self.dataset_filter_path = os.path.splitext(path)[0] + '_filters.json'
        self.dataset_filter = {}
        if os.path.exists(self.dataset_filter_path) and os.path.getsize(self.dataset_filter_path) > 0:
            with open(self.dataset_filter_path, 'rt') as f:
                self.dataset_filter.update(json.load(f))
        self.meta = create_dataset_meta(self.path)

    def server(self):
        return dict(canWrite=True)

    def user(self, email):
        return {}

    def datasets(self, email):
        results = []
        results.append(self.meta)
        return results

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        result = Entity(dataset_id, self.meta)
        return result

    def dataset_filters(self, email, dataset_id):
        results = []
        for key in self.dataset_filter:
            r = self.dataset_filter[key]
            r['id'] = key
            results.append(r)
        return results

    def delete_dataset_filter(self, email, filter_id):
        del self.dataset_filter[filter_id]
        self.__write_dataset_filter()

    def get_dataset_filter(self, email, filter_id):
        return self.dataset_filter[filter_id]

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        if filter_id is None:
            import uuid
            filter_id = str(uuid.uuid4())

        entity = self.dataset_filter.get(filter_id)
        if entity is None:
            entity = {}
            self.dataset_filter[filter_id] = entity
        if filter_name is not None:
            entity['name'] = filter_name
        if dataset_filter is not None:
            entity['value'] = json.dumps(dataset_filter)
        if email is not None:
            entity['email'] = email
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        if filter_notes is not None:
            entity['notes'] = filter_notes
        self.__write_dataset_filter()
        return filter_id

    def __write_dataset_filter(self):
        with open(self.dataset_filter_path, 'wt') as f:
            json.dump(self.dataset_filter, f)
